"""Test distributions.py functionality."""
import numpy as np
import tensorflow as tf
from scipy.linalg import cho_solve
from scipy.stats import wishart
from tensorflow.contrib.distributions import MultivariateNormalTriL

from aboleth.distributions import kl_sum, _chollogdet
from .conftest import SEED


def test_kl_normal_normal():
    """Test Normal/Normal KL."""
    dim = (5, 10)
    mu = np.zeros(dim, dtype=np.float32)
    std = 1.0

    q = tf.distributions.Normal(mu, std)

    # Test 0 KL
    p = tf.distributions.Normal(mu, std)
    KL0 = kl_sum(q, p)

    # Test diff var
    std1 = 2.0
    p = tf.distributions.Normal(mu, std1)
    KL1 = kl_sum(q, p)
    rKL1 = 0.5 * ((std / std1)**2 - 1 + np.log((std1 / std)**2)) * np.prod(dim)

    # Test diff mu
    mu1 = np.ones(dim, dtype=np.float32)
    p = tf.distributions.Normal(mu1, std)
    KL2 = kl_sum(q, p)
    rKL2 = 0.5 * (np.sum((mu1 - mu)**2) / std**2)

    tc = tf.test.TestCase()
    with tc.test_session():
        kl0 = KL0.eval()
        assert np.isscalar(kl0)
        assert kl0 == 0.
        assert np.allclose(KL1.eval(), rKL1)
        assert np.allclose(KL2.eval(), rKL2)


def test_kl_gaussian_normal(random):
    """Test Gaussian/Normal KL."""
    dim = (5, 10)
    Dim = (5, 10, 10)

    mu0 = random.randn(*dim).astype(np.float32)
    L0 = random_chol(Dim)
    q = MultivariateNormalTriL(mu0, L0)

    mu1 = random.randn(*dim).astype(np.float32)
    std1 = 1.0
    L1 = [(std1 * np.eye(dim[1])).astype(np.float32) for _ in range(dim[0])]
    p = tf.distributions.Normal(mu1, std1)

    KL = kl_sum(q, p)
    KLr = KLdiv(mu0, L0, mu1, L1)

    tc = tf.test.TestCase()
    with tc.test_session():
        kl = KL.eval()
        assert np.isscalar(kl)
        assert np.allclose(kl, KLr)


def test_kl_gaussian_gaussian(random):
    """Test Gaussian/Gaussian KL."""
    dim = (5, 10)
    Dim = (5, 10, 10)

    mu0 = random.randn(*dim).astype(np.float32)
    L0 = random_chol(Dim)
    q = MultivariateNormalTriL(mu0, L0)

    mu1 = random.randn(*dim).astype(np.float32)
    L1 = random_chol(Dim)
    p = MultivariateNormalTriL(mu1, L1)

    KL = kl_sum(q, p)
    KLr = KLdiv(mu0, L0, mu1, L1)

    tc = tf.test.TestCase()
    with tc.test_session():
        assert np.allclose(KL.eval(), KLr)


def test_chollogdet():
    """Test log det with cholesky matrices."""
    Dim = (5, 10, 10)
    L = random_chol(Dim)
    rlogdet = np.sum([logdet(l) for l in L])
    tlogdet = _chollogdet(L)

    L[0, 0, 0] = 1e-17  # Near zero to test numerics
    L[1, 3, 3] = -1.
    L[4, 5, 5] = -20.

    nlogdet = _chollogdet(L)

    tc = tf.test.TestCase()
    with tc.test_session():
        assert np.allclose(tlogdet.eval(), rlogdet)
        assert not np.isnan(nlogdet.eval())


def random_chol(dim):
    """Generate random pos def matrices."""
    D = dim[1]
    n = dim[0]
    np.random.seed(SEED)
    C = wishart.rvs(df=D, scale=10 * np.eye(D), size=n)
    np.random.seed(None)
    L = np.array([np.linalg.cholesky(c).astype(np.float32) for c in C])
    return L


def KLdiv(mu0, Lcov0, mu1, Lcov1):
    """Numpy KL calculation."""
    tr, dist, ldet = 0., 0., 0.
    D, n = mu0.shape
    for m0, m1, L0, L1 in zip(mu0, mu1, Lcov0, Lcov1):
        tr += np.trace(cho_solve((L1, True), L0.dot(L0.T)))
        md = m1 - m0
        dist += md.dot(cho_solve((L1, True), md))
        ldet += logdet(L1) - logdet(L0)

    KL = 0.5 * (tr + dist + ldet - D * n)
    return KL


def logdet(L):
    """Log Determinant from Cholesky."""
    return 2. * np.log(L.diagonal()).sum()



"""Test the layers module."""

import numpy as np
import tensorflow as tf

import aboleth as ab
from aboleth.losses import _sum_likelihood


def test_elbo_likelihood(make_graph):
    """Test for expected output dimensions from loss ."""
    x, y, N, X_, Y_, N_, layers = make_graph
    nn, kl = layers(X=X_)
    like = tf.distributions.Normal(nn, scale=1.)

    loss = ab.elbo(like, Y_, N, kl)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()

        L = loss.eval(feed_dict={X_: x, Y_: y, N_: float(N)})
        assert np.isscalar(L)


def test_map_likelihood(make_graph):
    """Test for expected output dimensions from deepnet."""
    x, y, _, _, _, _, layers = make_graph
    nn, reg = layers(X=x.astype(np.float32))
    like = tf.distributions.Normal(nn, scale=1.)

    loss = ab.max_posterior(like, y.astype(np.float32), reg)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()

        L = loss.eval()
        assert np.isscalar(L)


def test_categorical_likelihood(make_data):
    """Test aboleth with a tf.distributions.Categorical likelihood.

    Since it is a bit of an odd half-multivariate case.
    """
    x, y, _, = make_data
    N, K = x.shape

    # Make two classes (K = 2)
    Y = np.zeros(len(y), dtype=np.int32)
    Y[y[:, 0] > 0] = 1

    layers = ab.stack(
        ab.InputLayer(name='X', n_samples=10),
        lambda X: (X, 0.0)   # Mock a sampling layer, with 2-class output
    )

    nn, reg = layers(X=x.astype(np.float32))
    like = tf.distributions.Categorical(logits=nn)

    ELBO = ab.elbo(like, Y, N, reg)
    MAP = ab.max_posterior(like, Y, reg)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()

        assert like.probs.eval().shape == (10, N, K)
        assert like.prob(Y).eval().shape == (10, N)

        L = ELBO.eval()
        assert np.isscalar(L)

        L = MAP.eval()
        assert np.isscalar(L)


def test_sum_likelihood():
    """Test we can do weighted sums of likelihoods."""
    N = 5
    Net = np.ones(N, dtype=np.float32) * .5
    like = tf.distributions.Bernoulli(probs=Net)
    Y = np.ones(N)

    def weight_fn(Y):
        return Y * np.arange(N)

    unweighted = _sum_likelihood(like, Y, None)
    value = _sum_likelihood(like, Y, np.arange(N))
    call = _sum_likelihood(like, Y, weight_fn)

    tc = tf.test.TestCase()
    with tc.test_session():
        sumll = unweighted.eval()
        assert np.allclose(sumll, np.log(0.5) * N)

        sumll = value.eval()
        assert np.allclose(sumll, np.sum(np.log(0.5) * np.arange(N)))

        sumll = call.eval()
        assert np.allclose(sumll, np.sum(np.log(0.5) * np.arange(N)))



"""Random generators and state."""
import numpy as np


class SeedGenerator:
    r"""Make new random seeds deterministically from a base random seed."""

    def __init__(self):
        """Construct a SeedGenerator object."""
        self.state = np.random.RandomState()

    def set_hyperseed(self, hs):
        """Set the random seed state in this object.

        Parameters
        ----------
        hs : None, int, array_like
            seed the random state of this object, see numpy.random.RandomState
            for valid inputs.
        """
        self.state.seed(hs)

    def next(self):
        """Generate a random int using this object's base state.

        Returns
        -------
        result : int
            an integer that can be used to seed other random states
            deterministically.
        """
        result = self.state.randint(0, 2**32)
        return result

    def __next__(self):
        """Next generator."""
        return self.next()


# Dont judge me -- most RNGs are global vars
seedgen = SeedGenerator()


def set_hyperseed(hs):
    r"""Set the global hyperseed from which to generate all other seeds.

    Parameters
    ----------
    hs : None, int, array_like
        seed the random state of the global hyperseed, see
        numpy.random.RandomState for valid inputs.
    """
    seedgen.set_hyperseed(hs)


def endless_permutations(N):
    r"""
    Generate an endless sequence of permutations of the set [0, ..., N).

    If we call this N times, we will sweep through the entire set without
    replacement, on the (N+1)th call a new permutation will be created, etc.

    Parameters
    ----------
    N: int
        the length of the set

    Yields
    ------
    int :
        yeilds a random int from the set [0, ..., N)

    Examples
    -------
    >>> perm = endless_permutations(5)
    >>> type(perm)
    <class 'generator'>
    >>> p = next(perm)
    >>> p < 5
    True
    >>> p2 = next(perm)
    >>> p2 != p
    True
    """
    generator = np.random.RandomState(next(seedgen))

    while True:
        batch_inds = generator.permutation(N)
        for b in batch_inds:
            yield b
